(function() {
  var Action, Call, Client, Command, CommandDispatcher, Event, Get, PersistentCommand, Property, Q, Set, Watch, lines, recon;
  var __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; }, __slice = Array.prototype.slice, __hasProp = Object.prototype.hasOwnProperty, __extends = function(child, parent) {
    for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; }
    function ctor() { this.constructor = child; }
    ctor.prototype = parent.prototype;
    child.prototype = new ctor;
    child.__super__ = parent.prototype;
    return child;
  };
  Q = require("q");
  recon = require("recon");
  lines = require("lines-adapter");
  CommandDispatcher = (function() {
    function CommandDispatcher() {
      this.call = __bind(this.call, this);
      this.on_multi = __bind(this.on_multi, this);
      this.on = __bind(this.on, this);      this.commands = {};
    }
    CommandDispatcher.prototype.on = function(command, callback) {
      var add;
      add = function(commands, cmd) {
        var first, rest;
        first = cmd[0], rest = 2 <= cmd.length ? __slice.call(cmd, 1) : [];
        if (rest.length === 0) {
          return commands[first] = callback;
        } else {
          if (commands[first] === void 0) {
            commands[first] = {};
          }
          return add(commands[first], rest);
        }
      };
      return add(this.commands, command);
    };
    CommandDispatcher.prototype.on_multi = function(command, callback) {
      var add;
      add = function(commands, cmd) {
        var first, rest;
        first = cmd[0], rest = 2 <= cmd.length ? __slice.call(cmd, 1) : [];
        if (rest.length === 0) {
          if (commands[first] != null) {
            if (!(commands[first].length != null)) {
              throw new Error("Inconsistent command handlers.");
            }
            return commands[first].push(callback);
          } else {
            return commands[first] = [callback];
          }
        } else {
          if (commands[first] === void 0) {
            commands[first] = {};
          }
          return add(commands[first], rest);
        }
      };
      return add(this.commands, command);
    };
    CommandDispatcher.prototype.call = function(command) {
      var sub_call;
      sub_call = function(commands, cmd) {
        var first, handler, rest, _i, _len, _ref;
        if (cmd.length === 0) {
          return console.log("Not enough arguments in command " + command);
        } else {
          first = cmd[0], rest = 2 <= cmd.length ? __slice.call(cmd, 1) : [];
          switch (typeof commands[first]) {
            case "function":
              return commands[first].apply(commands, rest);
            case "object":
              if (commands[first] instanceof Array) {
                _ref = commands[first];
                for (_i = 0, _len = _ref.length; _i < _len; _i++) {
                  handler = _ref[_i];
                  handler.apply(null, rest);
                }
                return null;
              } else {
                return sub_call(commands[first], rest);
              }
          }
        }
      };
      return sub_call(this.commands, command);
    };
    return CommandDispatcher;
  })();
  PersistentCommand = (function() {
    function PersistentCommand() {
      this.add = __bind(this.add, this);
    }
    PersistentCommand.prototype.add = function() {};
    return PersistentCommand;
  })();
  Action = (function() {
    __extends(Action, PersistentCommand);
    function Action(client, path, callback) {
      this.client = client;
      this.path = path;
      this.callback = callback;
      this.add = __bind(this.add, this);
      this.client.dispatch.on(["docall", this.path], __bind(function() {
        var args;
        args = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
        return this.callback.apply(this, args);
      }, this));
    }
    Action.prototype.add = function() {
      return this.client.command("mkaction", this.path);
    };
    return Action;
  })();
  Event = (function() {
    __extends(Event, PersistentCommand);
    function Event(client, path) {
      this.client = client;
      this.path = path;
      this.raise = __bind(this.raise, this);
      this.add = __bind(this.add, this);
    }
    Event.prototype.add = function() {
      return this.client.command("mkevent", this.path);
    };
    Event.prototype.raise = function() {
      var args, _ref;
      args = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
      return (_ref = this.client).command.apply(_ref, ["raise", this.path].concat(__slice.call(args)));
    };
    return Event;
  })();
  Property = (function() {
    __extends(Property, PersistentCommand);
    function Property(client, path, getter, setter) {
      this.client = client;
      this.path = path;
      this.getter = getter;
      this.setter = setter;
      this.add = __bind(this.add, this);
      this.client.dispatch.on(["getprop", this.path], __bind(function() {
        return this.getter();
      }, this));
      this.client.dispatch.on(["setprop", this.path], __bind(function(value) {
        return this.setter(value);
      }, this));
    }
    Property.prototype.add = function() {
      return this.client.command("mkprop", this.path);
    };
    return Property;
  })();
  Watch = (function() {
    __extends(Watch, PersistentCommand);
    function Watch(client, path, callback) {
      this.client = client;
      this.path = path;
      this.callback = callback;
      this.add = __bind(this.add, this);
      this.client.dispatch.on_multi(["event", this.path], __bind(function() {
        var args;
        args = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
        return this.callback.apply(this, args);
      }, this));
    }
    Watch.prototype.add = function() {
      return this.client.command("watch", this.path);
    };
    return Watch;
  })();
  Command = (function() {
    function Command(client) {
      this.client = client;
      this.get_value = __bind(this.get_value, this);
      this.run = __bind(this.run, this);
      this.value = Q.defer();
    }
    Command.prototype.run = function() {};
    Command.prototype.get_value = function() {
      return this.value.promise;
    };
    return Command;
  })();
  Call = (function() {
    __extends(Call, Command);
    function Call() {
      var args, client, path;
      client = arguments[0], path = arguments[1], args = 3 <= arguments.length ? __slice.call(arguments, 2) : [];
      this.path = path;
      this.args = args;
      this.run = __bind(this.run, this);
      Call.__super__.constructor.call(this, client);
    }
    Call.prototype.run = function() {
      var _ref;
      return this.value.resolve((_ref = this.client).command.apply(_ref, ["call", this.path].concat(__slice.call(this.args))));
    };
    return Call;
  })();
  Set = (function() {
    __extends(Set, Command);
    function Set(client, path, new_value) {
      this.path = path;
      this.new_value = new_value;
      this.run = __bind(this.run, this);
      Set.__super__.constructor.call(this, client);
    }
    Set.prototype.run = function() {
      return this.value.resolve(this.client.command("set", this.path, this.new_value));
    };
    return Set;
  })();
  Get = (function() {
    __extends(Get, Command);
    function Get(client, path) {
      this.path = path;
      this.run = __bind(this.run, this);
      Get.__super__.constructor.call(this, client);
    }
    Get.prototype.run = function() {
      return this.value.resolve(this.client.command("get", this.path));
    };
    return Get;
  })();
  Client = (function() {
    function Client() {
      this.command = __bind(this.command, this);
      this.do_return = __bind(this.do_return, this);
      this.send_command = __bind(this.send_command, this);
      this.process_line = __bind(this.process_line, this);
      var cmd, on_connect, on_disconnect, shet_host, shet_port, _i, _len, _ref;
      this.return_callbacks = {};
      this.persistent_commands = [];
      this.waiting_commands = [];
      this.connected = false;
      this.dispatch = new CommandDispatcher;
      this.next_id = 0;
      _ref = Client.command_names;
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        cmd = _ref[_i];
        this[cmd] = this[cmd].bind(this);
      }
      on_connect = __bind(function() {
        var cmd, node, _j, _k, _len2, _len3, _ref2, _ref3;
        this.connected = true;
        _ref2 = this.persistent_commands;
        for (_j = 0, _len2 = _ref2.length; _j < _len2; _j++) {
          node = _ref2[_j];
          node.add();
        }
        _ref3 = this.waiting_commands;
        for (_k = 0, _len3 = _ref3.length; _k < _len3; _k++) {
          cmd = _ref3[_k];
          cmd.run();
        }
        return this.waiting_commands = [];
      }, this);
      on_disconnect = __bind(function() {
        return this.connected = false;
      }, this);
      shet_port = parseInt(process.env["SHET_PORT"]) || 11235;
      shet_host = process.env["SHET_HOST"] || "localhost";
      this.connection = recon(shet_host, shet_port);
      this.connection.on("connect", on_connect);
      this.connection.on("reconnect", on_connect);
      this.connection.on("drop", on_disconnect);
      lines(this.connection).on("data", this.process_line);
    }
    Client.command_names = [];
    Client.add_presistent_command = function(name, cls) {
      this.command_names.push(name);
      return this.prototype[name] = function() {
        var args, cmd;
        args = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
        cmd = (function(func, args, ctor) {
          ctor.prototype = func.prototype;
          var child = new ctor, result = func.apply(child, args);
          return typeof result === "object" ? result : child;
        })(cls, [this].concat(__slice.call(args)), function() {});
        this.persistent_commands.push(cmd);
        if (this.connected) {
          cmd.add();
        }
        return cmd;
      };
    };
    Client.add_command = function(name, cls) {
      this.command_names.push(name);
      return this.prototype[name] = function() {
        var args, cmd;
        args = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
        cmd = (function(func, args, ctor) {
          ctor.prototype = func.prototype;
          var child = new ctor, result = func.apply(child, args);
          return typeof result === "object" ? result : child;
        })(cls, [this].concat(__slice.call(args)), function() {});
        if (this.connected) {
          cmd.run();
        } else {
          this.waiting_commands.push(cmd);
        }
        return cmd.get_value();
      };
    };
    Client.add_presistent_command("add_action", Action);
    Client.add_presistent_command("add_event", Event);
    Client.add_presistent_command("watch", Watch);
    Client.add_presistent_command("add_prop", Property);
    Client.add_command("call", Call);
    Client.add_command("get", Get);
    Client.add_command("set", Set);
    Client.prototype.process_line = function(line) {
      var args, command, id, status, value, _ref, _ref2;
      _ref = JSON.parse(line), id = _ref[0], command = _ref[1], args = 3 <= _ref.length ? __slice.call(_ref, 2) : [];
      if (command === "return") {
        status = args[0], value = args[1];
        if (status === 0) {
          this.return_callbacks[id].resolve(value);
        } else {
          this.return_callbacks[id].reject(value);
        }
        return delete this.return_callbacks[id];
      } else {
        return Q.when((_ref2 = this.dispatch).call.apply(_ref2, [command].concat(__slice.call(args))), __bind(function(value) {
          return this.do_return(id, 0, value);
        }, this), __bind(function(value) {
          return this.do_return(id, 1, value);
        }, this));
      }
    };
    Client.prototype.send_command = function(cmd) {
      return this.connection.write((JSON.stringify(cmd)) + "\r\n");
    };
    Client.prototype.do_return = function(id, status, value) {
      return this.send_command([id, "return", status, value]);
    };
    Client.prototype.command = function() {
      var args, d, id, name;
      name = arguments[0], args = 2 <= arguments.length ? __slice.call(arguments, 1) : [];
      id = this.next_id++;
      this.send_command([id, name].concat(__slice.call(args)));
      d = Q.defer();
      this.return_callbacks[id] = d;
      return d.promise;
    };
    return Client;
  })();
  exports.Client = Client;
}).call(this);
